//------------------------------------------------
// Polled I2C routines
//------------------------------------------------
// Simple version without internal device address
// and not suitable for multimaster mode!
// WinAVR-20050214 & avr.libc-1.4.0 
//------------------------------------------------
// V-0.1 (draft, not tested) 21.11.2005
// eMail: peter.sager@freesurf.ch
//------------------------------------------------

//------------------------------------------------
// Includes
//------------------------------------------------

// AVR.LibC specific
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/twi.h>

// AVR-GCC specific
#include <stdint.h>
#include <stdio.h>

//------------------------------------------------
// Definition of my types
//------------------------------------------------

typedef signed char        s08;
typedef signed int         s16;
typedef signed long        s32;
typedef signed long long   s64;
typedef unsigned char      u08;
typedef unsigned int       u16;
typedef unsigned long      u32;
typedef unsigned long long u64;

//-------------------------------------------
// Definition of constants
//-------------------------------------------

#define TW_MT_ACK (TW_MT_SLA_ACK | TW_MT_DATA_ACK)  // MT-Acknowledge for address+data

#define F_SCL   100000UL  // I2C-Clock SCL=100 kHz
#define SDA     PC4       // IO for I2C-SDA (just as info)
#define SCL     PC5       // IO for I2C-SCL (just as info)

#define ADR_EXP1  0x40    // I2C-Adress for 8Bit I/O Expander PCF8574
#define ADR_EXP2  0x42    // I2C-Adress for 8Bit I/O Expander PCF8574

//-------------------------------------------
// declaration of function prototypes
//-------------------------------------------

void TWI_init(void);
void TWI_read (u08 TWI_adr, u08 *val, u08 count); 
void TWI_write(u08 TWI_adr, u08 *val, u08 count); 

//-------------------------------------------
// main:
//-------------------------------------------

int main(void)
{
  u08 temp;
  //-------------------------------------------
  // Set all PortC pins as Inuts with pullups
	//-------------------------------------------
  PORTC = 0xFF; // PPPP'PPPP	
	DDRC  = 0x00; // IIII'IIII
  //-------------------------------------------
  TWI_init();
  while(1)
  {
    TWI_read (ADR_EXP1, &temp, 1);   // read byte from  PCF8574(1)
    TWI_write(ADR_EXP2, &temp, 1);   // and write it to PCF8574(2)
  }
}

//-------------------------------------------
// functions
//-------------------------------------------


//--------------------------------------------------------
// Intialize the I2C-Bus Interface 
// as Master with SCL 100kHz, non interrupt driven
//--------------------------------------------------------

void TWI_init(void)
{
	//--------------------------------------------------------
	// SCL frequency: 102.400 kHz (F_CPU = 12 MHz)
	// SCL = F_CPU/(16+2*TWBR*4^TWSR)
	//--------------------------------------------------------
	TWBR = (u08)(F_CPU/2/F_SCL-8); // TWI-datrate divider = 52
	TWSR = 0x00;                     // TWI-clock prescaler = 1
	TWCR = (1<<TWEA)|(1<<TWEN);      // no interrupt mode!
}


void TWI_read(u08 TWI_adr, u08 *val, u08 count)
//------------------------------------------------
// Read n=count bytes from I2C-Bus in polled mode
// needs about 200us for reading addr + one byte
//------------------------------------------------
{
  u08 i=0;
  TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);     // send START condition
  while (!(TWCR & (1<<TWINT)));               // wait for response
  if ((TWSR & 0xF8) == TW_START)              // if START was successful
  {
    TWDR = TWI_adr | 0x01;                    // put slave address+READ to TWDR
    TWCR = (1<<TWINT)|(1<<TWEN);              // and start transmission
    while (!(TWCR & (1<<TWINT)));             // wait for response
    if ((TWSR & 0xF8) == TW_MR_SLA_ACK)       // if acknowledge ok
    {
      count--;
      while (count>i)                         // while more than one byte to read
      {
        TWCR=(1<<TWINT)|(1<<TWEA)|(1<<TWEN);  // then start transfer with acknowledge
        while (!(TWCR & (1<<TWINT)));         // wait for response
        val[i++]=TWDR;                        // read databyte from TWDR
      }
      TWCR=(1<<TWINT)|(1<<TWEN);              // start last byte transfer without acknowledge
      while (!(TWCR & (1<<TWINT)));           // wait for response
      val[i]=TWDR;                            // read databyte from TWDR
    }
    else
    {
      // UART0_putp("Error: I2C[R] Addr\r\n");
    }
  }
  else
  {
    // UART0_putp("Error: I2C[R] Start\r\n");
  }
  TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO);   // transmit STOP condition
}



void TWI_write(u08 TWI_adr, u08 *val, u08 count)
//------------------------------------------------
// write n=count bytes to I2C-Bus in polled mode
// needs about 200us for writing Addr + one byte
//------------------------------------------------
{
  TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);   // send START condition
  while (!(TWCR & (1<<TWINT)));             // wait for response
  if ((TWSR & 0xF8) == TW_START)            // if START was successful
  {
    TWDR = (TWI_adr & 0xF0);                // put slave address+WRITE to TWDR
    TWCR = (1<<TWINT)|(1<<TWEN);            // and start transmission
    for (u08 i=0; i<count; i++)
    {
      while (!(TWCR & (1<<TWINT)));         // wait for response
      if ((TWSR & 0xC8) == TW_MT_ACK)       // if acknowledge ok
      {
        TWDR = val[i];                      // put databyte  to TWDR
        TWCR = (1<<TWINT)|(1<<TWEN);        // and start transmission
      }
      else
      {
        // UART0_putp("Error: I2C[W] Data\r\n");
      }
    }
  }
  else
  {
    // UART0_putp("Error: I2C[W] Start\r\n");
  }
  TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO);   // transmit STOP condition
}
